"""Timestamp writer for keeping track of bulk optimizations."""

from treestamps.tree.dump import TreestampsDump
from treestamps.tree.load import TreestampLoad
from treestamps.tree.set import TreestampsSet


class Treestamps(TreestampsSet, TreestampLoad, TreestampsDump):
    """Treestamps object to hold settings and caches."""



"""Tree config."""

from dataclasses import dataclass
from pathlib import Path

from treestamps.config import CommonConfig

_CONFIG_KEYS = ("ignore", "symlinks")


@dataclass
class TreestampsConfig(CommonConfig):
    """Config data."""

    path: Path = Path()

    def __post_init__(self):
        """Fix types."""
        super().__post_init__()
        self.path = Path(self.path)

    def get_config_dict(self):
        """Return select attributes as a dict."""
        result = {}
        for config_key in _CONFIG_KEYS:
            result[config_key] = getattr(self, config_key)
        return result



"""Treestamps Config methods."""

from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from types import MappingProxyType


@dataclass
class CommonConfig:
    """Common Config meant to be subclassed."""

    program_name: str
    verbose: int = 0
    symlinks: bool = True
    ignore: Iterable[str] = frozenset()
    check_config: bool = True
    program_config: Mapping | None = None
    program_config_keys: Iterable[str] = frozenset()

    @classmethod
    def _normalize_config(cls, config: Mapping | None) -> MappingProxyType | None:
        """Recursively convert iterables into sorted unique lists."""
        if config is None:
            return None
        new_config: dict = {}
        for key, value in config.items():
            if isinstance(value, list | tuple | set | frozenset):
                new_config[key] = tuple(sorted(frozenset(value)))
            elif isinstance(value, Mapping):
                normalized = cls._normalize_config(value)
                if normalized is not None:
                    normalized = MappingProxyType(dict(sorted(normalized.items())))
                new_config[key] = normalized
            else:
                new_config[key] = value

        return MappingProxyType(new_config)

    def __post_init__(self):
        """Fix types and normalize program config dict."""
        self.ignore = frozenset(self.ignore)
        self.program_config_keys = frozenset(self.program_config_keys)

        # Filter dict by keys
        if self.program_config:

            def filter_func(pair):
                return pair[0] in self.program_config_keys

            self.program_config = dict(filter(filter_func, self.program_config.items()))

        self.program_config = self._normalize_config(self.program_config)



"""Load methods."""

from pathlib import Path
from types import MappingProxyType

from termcolor import cprint

from treestamps.tree.get import TreestampsGet


class TreestampLoad(TreestampsGet):
    """Load methods."""

    def _is_path_skipped(self, path: Path) -> bool:
        """Return if path is ignored or not allowed because symlink."""
        return any(path.match(ignore_glob) for ignore_glob in self._config.ignore) or (
            not self._config.symlinks and path.is_symlink()
        )

    def _load_timestamps_file_pop_config_matches(self, yaml):
        """Return if the configured and loaded configs match."""
        if not self._config.check_config:
            return True

        yaml_ts_config = yaml.pop(self._TREESTAMPS_CONFIG_TAG, {})
        # convert to internal type
        yaml_ts_config["ignore"] = frozenset(yaml_ts_config.get("ignore", []))
        if self._config.get_config_dict() != yaml_ts_config:
            return False

        yaml_program_config = yaml.pop(self._CONFIG_TAG, None)
        if yaml_program_config is not None:
            # convert to internal type
            yaml_program_config = MappingProxyType(yaml_program_config)
        return self._config.program_config == yaml_program_config

    def _get_absolute_entry_path(self, root: Path, path_str) -> Path:
        """Get the path from a timestamp file relative to this treestamp's root_dir."""
        abs_root = self._get_absolute_path(self.root_dir, root)
        if not abs_root:
            raise ValueError
        abs_path = self._get_absolute_path(abs_root, path_str)
        if not abs_path:
            raise ValueError
        return abs_path

    def _load_timestamp_entry(self, root: Path, path_str: str, ts: float) -> None:
        """Load a single timestamp entry into the cache."""
        try:
            abs_path = self._get_absolute_entry_path(root, path_str)
            old_ts = self.get(abs_path)
            if abs_path not in self._timestamps or old_ts is None or ts > old_ts:
                self._timestamps[abs_path] = ts
        except Exception as exc:
            cprint(f"WARNING: Invalid timestamp for {path_str}: {ts} {exc}", "yellow")

    def _load_timestamps_file(self, timestamps_path: Path) -> None:
        """Load timestamps from a file."""
        if not timestamps_path.is_file():
            return

        try:
            yaml = self._YAML.load(timestamps_path)
            if not yaml:
                return

            # Pop off config entries and compare.
            if not self._load_timestamps_file_pop_config_matches(yaml):
                return
            # Pop off the WAL
            wal = yaml.pop(self._WAL_TAG, ())

            # What's left are timestamps
            entries = dict(yaml)

            # Update entries with WAL entries
            for wal_entry in wal:
                try:
                    entries.update(wal_entry)
                except Exception as exc:
                    cprint(f"WARNING: loading WAL entry: {wal_entry}: {exc}", "yellow")

            for path_str, ts in entries.items():
                self._load_timestamp_entry(timestamps_path.parent, path_str, ts)
        except Exception as exc:
            cprint(f"ERROR: parsing timestamps file: {timestamps_path} {exc}", "red")

    def _consume_child_timestamps(self, path: Path) -> None:
        """Consume a child timestamp and add its values to our root."""
        try:
            if not path.is_file() or self._is_path_skipped(path):
                return
            self._load_timestamps_file(path)
            if path != self._dump_path:
                self._consumed_paths.add(path)
            if self._config.verbose:
                cprint(f"Read timestamps from {path}")
        except Exception as exc:
            cprint(f"WARNING: reading child timestamps {exc}", "yellow")

    def _consume_all_child_timestamps(
        self, path: Path, do_consume_children: bool
    ) -> None:
        """Recursively consume all timestamps and wal files."""
        try:
            if not path.is_dir() or self._is_path_skipped(path):
                return
            for name in (self._filename, self._wal_filename):
                self._consume_child_timestamps(path / name)
            if do_consume_children:
                for dir_entry in path.iterdir():
                    self._consume_all_child_timestamps(
                        dir_entry, do_consume_children=True
                    )
        except Exception as exc:
            cprint(f"WARNING: reading all child timestamps {exc}", "yellow")

    def _load_parent_timestamps(self, path: Path) -> None:
        """Recursively load timestamps from all parents."""
        if path.parent == path.parent.parent or self._is_path_skipped(path):
            return
        parent = path.parent
        self._load_timestamps_file(parent / self._filename)
        self._load_timestamps_file(parent / self._wal_filename)
        self._load_parent_timestamps(parent)

    def load(self) -> None:
        """Load all timestamps."""
        self._load_parent_timestamps(self.root_dir)
        do_consume_children = self._config.path.is_dir()
        self._consume_all_child_timestamps(self.root_dir, do_consume_children)



"""Common Writing methods."""

from contextlib import suppress
from pathlib import Path

from treestamps.tree.init import TreestampsInit


class TreestampsWrite(TreestampsInit):
    """Common Writing methods."""

    def _close_wal(self) -> None:
        """Close the write ahead log."""
        if self._wal is None:
            return
        with suppress(AttributeError):
            self._wal.close()
        self._wal = None

    def _get_relative_path_str(self, abs_path: Path) -> str:
        """Get the path string relative to the root_dir."""
        return str(abs_path.relative_to(self.root_dir))

    def _get_dumpable_program_config(self) -> dict:
        """Set the config tag in the yaml to be dumped."""
        # NOTE: Treestamps symlinks & ignore options should be represented in
        # the program config.
        yaml = {}
        if self._config.program_config is not None:
            yaml[self._CONFIG_TAG] = dict(self._config.program_config)
        yaml[self._TREESTAMPS_CONFIG_TAG] = {
            "ignore": self._config.ignore,
            "symlinks": self._config.symlinks,
        }
        return yaml

    def _dump_to_file(self, path, yaml):
        """Dump to file."""
        config_yaml = self._get_dumpable_program_config()
        yaml.update(config_yaml)
        self._close_wal()
        self._YAML.dump(yaml, path)



"""Test classmethods."""

from tests import PROGRAM
from tests.integration.base_test import BaseTestDir
from treestamps.grove import Grovestamps, GrovestampsConfig

__all__ = ()

PROGRAM_NAME = f"{PROGRAM}-tests"


class TestCycle(BaseTestDir):
    """Test classmethods."""

    def _dump(self, config, subpaths):
        gs = Grovestamps(config)

        times = {}
        for subpath in subpaths:
            ts = gs.get(subpath)
            if not ts:
                print("No stamp from", subpath)
            assert ts

            names = (
                "testname",
                "test: name",
                "test/name",
                "Edgar Rice Burroughs/Tarzan of the Apes: Three Complete Nivels.epub",
            )

            for name in names:
                path = subpath / name
                times[path] = ts.set(path)

        gs.dump()
        return times

    def _load(self, config, subpaths, times):
        print(config)
        gs = Grovestamps(config)

        for subpath in subpaths:
            ts = gs.get(subpath)
            if not ts:
                print("No stamp from", subpath, ts)
            assert ts

            for path, stamp in times.items():
                if not path.is_relative_to(subpath):
                    continue
                loaded_stamp = ts.get(path)
                if stamp != loaded_stamp:
                    print("stamp mismatch:", stamp, "!=", loaded_stamp)
                    print(path, "ts._timestamps", ts._timestamps)
                assert stamp == loaded_stamp

    def test_set_dump_load_get(self):
        """Test it all."""
        # Make subdirs
        subdirs = ("a", "b", "c")
        subpaths = []
        for subdir in subdirs:
            subpath = self.TMP_ROOT / subdir
            subpath.mkdir()
            subpaths.append(subpath)

        config = GrovestampsConfig(PROGRAM_NAME, paths=subpaths, verbose=10)

        times = self._dump(config, subpaths)

        for subpath in subpaths:
            stamp_path = subpath / f".{PROGRAM_NAME}_treestamps.yaml"
            assert stamp_path.exists()

        self._load(config, subpaths, times)
